@@ -24,7 +24,7 @@ from types import ModuleType
 from typing import Callable, Iterable, List, Optional, Sequence, Set, Union, Dict
 
 import pkg_resources
-import regex as re
+import re
 
 PathStr = Union[Path, str]
 AnimFunc = Callable[[Optional[float]], 'OpenSCADObject']
@@ -611,54 +611,27 @@ def sp_code_in_scad_comment(calling_file: PathStr) -> str:
 # ===========
 # = Parsing =
 # ===========
-def extract_callable_signatures(scad_file_path: PathStr) -> List[dict]:
-    scad_code_str = Path(scad_file_path).read_text()
-    return parse_scad_callables(scad_code_str)
+def parse_scad_callables(filename: str) -> List[dict]:
+    from .py_scadparser import scad_parser
+
+    modules, functions, _ = scad_parser.parseFile(filename)
 
-def parse_scad_callables(scad_code_str: str) -> List[dict]:
     callables = []
-
-    # Note that this isn't comprehensive; tuples or nested data structures in
-    # a module definition will defeat it.
-
-    # Current implementation would throw an error if you tried to call a(x, y)
-    # since Python would expect a(x);  OpenSCAD itself ignores extra arguments,
-    # but that's not really preferable behavior
-
-    # TODO:  write a pyparsing grammar for OpenSCAD, or, even better, use the yacc parse grammar
-    # used by the language itself.  -ETJ 06 Feb 2011
-
-    # FIXME: OpenSCAD use/import includes top level variables. We should parse 
-    # those out (e.g. x = someValue;) as well -ETJ 21 May 2019
-    no_comments_re = r'(?mxs)(//.*?\n|/\*.*?\*/)'
-
-    # Also note: this accepts: 'module x(arg) =' and 'function y(arg) {', both
-    # of which are incorrect syntax
-    mod_re = r'(?mxs)^\s*(?:module|function)\s+(?P<callable_name>\w+)\s*\((?P<all_args>.*?)\)\s*(?:{|=)'
-
-    # See https://github.com/SolidCode/SolidPython/issues/95; Thanks to https://github.com/Torlos
-    args_re = r'(?mxs)(?P<arg_name>\w+)(?:\s*=\s*(?P<default_val>([\w.\"\s\?:\-+\\\/*]+|\((?>[^()]|(?2))*\)|\[(?>[^\[\]]|(?2))*\])+))?(?:,|$)'
-
-    # remove all comments from SCAD code
-    scad_code_str = re.sub(no_comments_re, '', scad_code_str)
-    # get all SCAD callables
-    mod_matches = re.finditer(mod_re, scad_code_str)
-
-    for m in mod_matches:
-        callable_name = m.group('callable_name')
+    for c in modules + functions:
         args = []
         kwargs = []
-        all_args = m.group('all_args')
-        if all_args:
-            arg_matches = re.finditer(args_re, all_args)
-            for am in arg_matches:
-                arg_name = am.group('arg_name')
-                # NOTE: OpenSCAD's arguments to all functions are effectively
-                # optional, in contrast to Python in which all args without
-                # default values are required.
-                kwargs.append(arg_name)
 
-        callables.append({'name': callable_name, 'args': args, 'kwargs': kwargs})
+        #for some reason solidpython needs to treat all openscad arguments as if
+        #they where optional. I don't know why, but at least to pass the tests
+        #it's neccessary to handle it like this !?!?!
+        for p in c.parameters:
+            kwargs.append(p.name)
+            #if p.optional:
+            #    kwargs.append(p.name)
+            #else:
+            #    args.append(p.name)
+
+        callables.append({'name': c.name, 'args': args, 'kwargs': kwargs})
 
     return callables
 
@@ -739,9 +712,23 @@ def new_openscad_class_str(class_name: str,
 def _subbed_keyword(keyword: str) -> str:
     """
     Append an underscore to any python reserved word.
+    Prepend an underscore to any OpenSCAD identifier starting with a digit.
     No-op for all other strings, e.g. 'or' => 'or_', 'other' => 'other'
     """
-    new_key = keyword + '_' if keyword in PYTHON_ONLY_RESERVED_WORDS else keyword
+    new_key = keyword
+
+    if keyword in PYTHON_ONLY_RESERVED_WORDS:
+        new_key = keyword + "_"
+
+    elif keyword[0].isdigit():
+        new_key = "_" + keyword
+
+    elif keyword == "$fn":
+        new_key = "segments"
+
+    elif keyword[0] == "$":
+        new_key = "__" + keyword[1:]
+
     if new_key != keyword:
         print(f"\nFound OpenSCAD code that's not compatible with Python. \n"
               f"Imported OpenSCAD code using `{keyword}` \n"
@@ -751,10 +738,22 @@ def _subbed_keyword(keyword: str) -> str:
 def _unsubbed_keyword(subbed_keyword: str) -> str:
     """
     Remove trailing underscore for already-subbed python reserved words.
+    Remove prepending underscore if remaining identifier starts with a digit.
     No-op for all other strings: e.g. 'or_' => 'or', 'other_' => 'other_'
     """
-    shortened = subbed_keyword[:-1]
-    return shortened if shortened in PYTHON_ONLY_RESERVED_WORDS else subbed_keyword
+    if subbed_keyword.endswith("_") and subbed_keyword[:-1] in PYTHON_ONLY_RESERVED_WORDS:
+        return subbed_keyword[:-1]
+
+    elif subbed_keyword.startswith("__"):
+        return "$" + subbed_keyword[2:]
+
+    elif subbed_keyword.startswith("_") and subbed_keyword[1].isdigit():
+        return subbed_keyword[1:]
+
+    elif subbed_keyword == "segments":
+        return "$fn"
+
+    return subbed_keyword
 
 # now that we have the base class defined, we can do a circular import
 from . import objects
